function [Bayes_pred_sets,MN_pred_sets,f_xbarh_X_g_save,like_bcd,post_bcd,sigma_save,ier_sigma_save] ... 
         = lr_pred_set(u,q,hfrac_vec,clevel_vec,bcd_mat,sigma_save,gdist,lfd_mat,cvalue_mat);
%This function calculates the long-run prediction in Mueller and Watson's
%Measuring Uncertainty about Long-Run Predictions
%The steps listed in the function correspond to the steps listed in
%Appendix of that paper
% ------------ Input Parameters
%   u:  Data vector ... missing values are not allowed
%            Dimension: (Tx1)        
%   q:  Number of low-frequency cosine transformations to use
%            Dimension: scalar
%   hfrac_vec: This vector contains the forecast horizons expressed as
%              fractions of T. Because this a vector, multiple horizons can
%              be computed with one call to the function
%            Dimension: (n_horizons x 1)
%   clevel_vec: This vector contains the "confidence" levels. that is the
%               coverage level of the prediction sets (e.g., 50%, 80%, 90%)
%            Dimension: (n_clevels x 1)
%   bcd_mat: This matrix contains the grid of values of 'bcd' ('theta' in the
%            paper) under consideration.
%             Dimension: (n_bcd x 1)
%   sigma_save:  This 3-dimensional matrix contains the "sigma" matrices for each row of bcd_mat 
%                and including all forecast horizons.
%                  If sigma_save = scalar, then the sigma matrices are
%                  computed by the program.
%                  Otherwise the dimension of sigma_mat is:
%                  Demension:(q+n_horizons)x(q_n_horizons)x(n_bcd);
%   gdist: This is (a) the Bayes prior for 'bcd'. It is ALSO the weighting
%          matrix for the set volume used in the frequentist calcuation. In
%          both cases this is called "Gamma" in the paper.
%              Dimesnion: n_bcd x 1
%   lfd_mat: This is a matrix of least favorable distributions (over bcd). 
%            As discussed in the paper, the LFD depends on q,"gdist',
%            horizon, and coverage level.  If lfd_mat = scalar, the MN
%            prediction sets are not computed.  Otherwise the lfd_mat is a
%            3 dimensional matrix with 
%              Dimension: (n_bcd) x (n_horizons) x (n_clevels)
%   cvalue_mat: This is a matrix of critical values for each of the least favorable 
%            distributions. If the lfd_mat is not given, then this is not used (so it 
%            can be an arbitrary scalar. Otherwise, its dimension is
%               Dimension: (n_horizons) x (n_clevels) 
% --------------- Output parameters
%
%   Bayes_pred_sets:  This contains the lower and upper endpoints of the
%                     equal-tail Bayes predictive sets.  The dimension is
%                        Dimension: 2 x (n_clevels) x (n_horizons)
%                     where (1,.,.) corresponds to the lower endpoint and
%                           (2,.,.) corresponds to the upper endpoint.
%   MN_pred_sets:  This contains the lower and upper endpoints of the
%                     equal-tail MN-predictive sets.  The dimension is
%                        Dimension: 2 x (n_clevels) x (n_horizons)
%                     where (1,.,.) corresponds to the lower endpoint and
%                           (2,.,.) corresponds to the upper endpoint. 
%                     Note: missing values are returned if LFD is not
%                     supplied to the function
%  f_xbarh_X_g_save:   This is the Bayes predictive density of xbar(T+1:T+h)
%                  given the sample data (as summarized by xbar(1:T) and
%                  X). The dimension is
%                     Dimension: (ngrid) x 2 x (n_horizons)
%                  where ngrid denotes a grid of values of xbar(T+1:T+h)
%                  where the predictive density is computed. The relevant
%                  entries are 
%                        (.,1,.): grid of values of xbar(T+1:T+h)
%                        ( .,2,.) value of predictive density
% like_bcd:        This contains the likelihood values for each value of
%                  bcd. The dimension is 
%                      Dimension: (n_bcd) x 1
% post_bcd:        This is the posterior distribution for the bcd values
%                       Dimension: (n_bcd) x 1
% sigma_save:      This returns the value of the matrix sigma_save.  If
%                  sigma_save is input, its value on the return is
%                  unaltered.  If sigma_save is not input, the value
%                  internally computed by the function is returned. (This
%                  can be useful for later calls to the function.
% ier_sigma_save:  This matrix contains error codes for the computation of
%                  the Sigma matrice for each value of bcd. 0 indicates no
%                  error. 1 indicates a non-pd Sigma, which is corrected by
%                  taking the absolute value of tbe offending eigenvalue. 
%                    Dimension: (n_bcd) x 1
%
%
% -- Parameter Names
  T = size(u,1);
  n_horizons = size(hfrac_vec,1);  % Number of Forecast Horizons
  n_clevels = size(clevel_vec,1);  % Number of "Confidence" levels
  n_bcd = size(bcd_mat,1);         % number of 'bcd' parameter values
  
   % Check for missing values, sizes of arrays and so forth
  if isnan(u) > 0;
      error('missing values in data vector');
  end;
  if size(gdist,1) ~= n_bcd;
      error('gdist has wrong dimension');
  end;
 
 %------------------ Step 1: Compute xbar(1:T), Cosine Transform and Standardised Cosine transform
 psi = psi_compute(T,q);
 xbar = mean(u);          % Sample Mean
 X = psi'*(u - xbar);     % Cosine Transformaion
 Xs = X/sqrt(X'*X);       % Normalize Cos Transformation
 
 %------------------- Step 2: Compute Covariance Matrix for each value of hfrac
 ier_sigma_save = zeros(n_bcd,1);
 if size(sigma_save,1) == 1;  % Compute sigma_save 
     sigma_save = NaN*zeros(q+n_horizons,q+n_horizons,n_bcd);
     for i = 1:n_bcd;
         b = bcd_mat(i,1);
         c = bcd_mat(i,2);
         d = bcd_mat(i,3);
         [sigma,ier] = Sigma_Compute(b,c,d,q,hfrac_vec);
         tmp = eig(sigma);
         sigma_save(:,:,i)=sigma;
         ier_sigma_save(i) = ier;
     end;
 end;
  
 % ------------------- Step 3: Bayes Analysis
   % 3.a: gdist is prior 
   % 3.b: Compute likelihood for each value of theta
   f_x = NaN*zeros(n_bcd,1);
   for i=1:n_bcd;
       sigma = sigma_save(1:q,1:q,i);
       sigma_inv = inv(sigma);
       f_x(i) = den_invariant(Xs,sigma_inv);
   end;
   like_bcd = f_x;                     % Likelihood values
   f_marg_x = gdist'*f_x;              % Marginal Likelihood
   post_bcd = (gdist.*f_x)/f_marg_x;   % Posterior
      
   % 3.c.i: Compute Predictive Density for y over a grid of values of y
     ngrid = 1001;                             % Number of grid points for predicive density
     f_y_x_g_save = NaN*zeros(ngrid,2,n_horizons);    % Predictive density for each horizon in terms of Ys
     f_xbarh_X_g_save = NaN*zeros(ngrid,2,n_horizons);% Predictive density for each horizon in terms of Xbar(T+1:T+h)
     f_yx_save = NaN*zeros(ngrid,n_bcd,n_horizons);   % Joint density for each horizon
     for ih = 1:n_horizons  % Compute prediction set for each horizon
         igrid = 0;            % Indicator for width of grid for Ys
         smin = -40/sqrt(q);   % Min and Max values of Ys; these are lengthened if too small
         smax = 40/sqrt(q);
         while igrid == 0;
             dd = (smax-smin)/(ngrid-1);
             Ys_grid = (smin:dd:smax)';   % grid of values for Ys
             Xs_Ys_grid = [repmat(Xs,1,size(Ys_grid,1)); Ys_grid'];   % grid of values for Xs and Ys (Xs held at sample values)
             f_yx = NaN*zeros(size(Ys_grid,1),n_bcd);  % for joint density
             for i = 1:n_bcd; 
                sigma = zeros(q+1,q+1);
                sigma(1:q,1:q) = sigma_save(1:q,1:q,i); 
                sigma(1:q,q+1) = sigma_save(1:q,q+ih,i);
                sigma(q+1,1:q) = sigma_save(q+ih,1:q,i);
                sigma(q+1,q+1) = sigma_save(q+ih,q+ih,i);
                sigma_inv = inv(sigma);
                f_yx(:,i) = den_invariant(Xs_Ys_grid,sigma_inv);
             end;
             % Compute Conditional Density
             f_y_x = f_yx./repmat(f_x',size(f_yx,1),1);  % Conditional density for each value of theta and Ys in grid
             f_y_x_g = f_y_x*post_bcd;     % Predictive Density using g-prior for year Ys in grid
             f_y_x_g_norm = f_y_x_g*dd;    % Predictive Density normalized so that grid over y sums to 1 (instead of integrates to 1)
             % Check that predictive density integrates to 1-small; otherwise increae grid size 
             if abs(sum(f_y_x_g_norm)-1) < .001;
                 igrid = 1;
             else;
                 fprintf('G-Bayes density integrates to %7.5f\n',sum(f_y_x_g_norm));
                 fprintf('Increasing smin and smax, ih = %4d, smin = %6.1f, smax = %6.1f\n',ih,smin,smax);
                 smin = smin-(5/sqrt(q));
                 smax = smax+(5/sqrt(q));
             end;
         end;
         f_yx_save(:,:,ih) = f_yx;                          % Save Joint Density
         f_y_x_g_save(:,:,ih) = [Ys_grid f_y_x_g];          % Save Predictive Density
         xbarh_grid = xbar + sqrt(X'*X)*Ys_grid;
         f_xbarh_X_g_save(:,:,ih) = [xbarh_grid f_y_x_g];   % Save Bayes Predictive Density
     end;
     
        % 3.c.iii: Compute Equal-Tail Prediction Sets for Ys
     Bayes_pred_sets_Ys = NaN*zeros(2,n_clevels,n_horizons);
     for ih = 1:n_horizons  % Compute prediction set for each horizon
         Ys_grid = f_y_x_g_save(:,1,ih);
         f_y_x_g = f_y_x_g_save(:,2,ih);
         dd = Ys_grid(2)-Ys_grid(1);
         f_y_x_g_norm = f_y_x_g*dd;
         for ic = 1:n_clevels;
             clevel=clevel_vec(ic);
             [cl,cu] = ci_f(f_y_x_g_norm,Ys_grid,clevel);
             Bayes_pred_sets_Ys(1,ic,ih) = cl;
             Bayes_pred_sets_Ys(2,ic,ih) = cu;
         end
     end;
     
     % 3.d.ii: Compute Equal-Tail Prediction Sets for u
     Bayes_pred_sets = xbar + sqrt(X'*X)*Bayes_pred_sets_Ys;
     

     
     
     % ------------------- Step 4: Construct MN-Frequentist Prediction Sets
      % Step 4.a:  The weighting function for theta is gdist
      % Step 4.b:  The Bayes sets were computed in step 3 for gdist
      % Step 4.c:  The LFDs and critical values are computed elsewhere and are input here
      % Step 4.d:  Compute the A(Ys) sets
      % -- Skip Step 4 if lfd is not provided 
      MN_pred_sets = NaN*zeros(2,n_clevels,n_horizons);
      if size(lfd_mat) > 1;
           A_pred_sets_Ys = NaN*zeros(2,n_clevels,n_horizons);
           for ih = 1:n_horizons  % Compute prediction set for each horizon
               Ys_grid = f_y_x_g_save(:,1,ih);
               f_yx = f_yx_save(:,:,ih);
               for ic = 1:n_clevels;
                  lamdist = lfd_mat(:,ih,ic);
                  cvalue = cvalue_mat(ih,ic);
                  f_yx_lam = f_yx*lamdist;
                  iin = f_yx_lam >= cvalue*f_marg_x;
                  if sum(iin) == 0;
                      cl_lfd = NaN;
                      cu_lfd = NaN;
                  else;
                      trend = (1:size(Ys_grid,1));
                      ind_include = trend(iin==1);  % Indices to include
                      ind_min = min(ind_include);
                      ind_max = max(ind_include);
                      cl_lfd = Ys_grid(ind_min);
                      cu_lfd = Ys_grid(ind_max);
                      % Check to see if this an interval
                      tmp = iin(ind_min:ind_max);
                      if sum(tmp) ~= (ind_max-ind_min+1);
                          fprintf('LFD CI is not an interval ... \n');
                          fprintf('    clevel = %4.2f\n',clevel_vec(ic));
                          fprintf('    horizon = %4.2f\n',hfrac_vec(ih));
                      end;
                  end;
                  A_pred_sets_Ys(1,ic,ih) = cl_lfd;
                  A_pred_sets_Ys(2,ic,ih) = cu_lfd; 
               end;
           end;
     
          % Step 4.e:  Compute the A(X) sets
          A_pred_sets = xbar + sqrt(X'*X)*A_pred_sets_Ys;
       
          % Step 4.f: Compute MN-Prediction sets
          for ih = 1:n_horizons;
              for ic = 1:n_clevels;
                 MN_pred_sets(1,ic,ih) = min(packr([Bayes_pred_sets(1,ic,ih);A_pred_sets(1,ic,ih)]));
                 MN_pred_sets(2,ic,ih) = max(packr([Bayes_pred_sets(2,ic,ih);A_pred_sets(2,ic,ih)]));
              end;
          end;
      end;               
end

